/*
   This program demonstrates a very subtle race condition
   that depends on the way the C code gets compiled and run.

   The line of C code
      count++;
   can be compiled two ways on an Intel CPU. One way to compile
   count++ is with the following i386 assembly code that uses
   an atomic memory access.
     incl  _count
   Another way to compile count++ is with this more RISC like,
   load-and-store sequence of three instructions.
      movl  _count, %eax
      incl  %eax
      movl  %eax, _count

   In the thread function in the code below, if count++ is compiled
   as the atomic instruction
     incl  _count
   and if all the threads are run on a single core, then the multiple
   threads do not have a race condition. (If the threads are run on
   multiple cores, then the fact that count++ is atomic doesn't matter,
   and there will be a race condition. Be sure you understand why.)

   One the other hand, in the thread function below, if count++ is
   compiled as
      movl  _count, %eax
      incl  %eax
      movl  %eax, _count
   then, even if we run the threads on a single core, we will probably
   get a race condition since the multiple instructions allow for the
   scheduler to take the CPU away from a thread in the middle of
   executing count++ (make sure you understand why that can lead to
   a data race condition)

   The GCC compiler used to have a compiler option, -fforce-mem, that
   told gcc to compile code in the more RISC way and another option,
   -fno-force-mem, that compiled code the other way. GCC now defaults
   to the RISC like code and this cannot be changed. But we can get the
   atomic count++ by using our own inline assembly code.

   The non-standard, gcc specific, line of C code
      asm("incl  _count");
   is compiled by gcc to
      incl  _count
   So by using this "inline assembly code", we can get an atomic count++.

   See the following page for more info about inline assembly.
      http://www.ibiblio.org/gferg/ldp/GCC-Inline-Assembly-HOWTO.html

   This example is based on Figure 13.16 badcnt.c, page 872, from
    "Computer Systems: A Programmers Perspective", by R Bryant and D. O'Hallaron.
*/
#include <stdio.h>    // printf()
#include "pthread.h"

#define ITERATIONS 100000000

unsigned count = 0; // global shared variable

int main( int argc, char *argv[])
{
   void *thread(void *vargp);  //thread function prototype

   #define NUMBER_OF_THREADS 6
   pthread_t tid[NUMBER_OF_THREADS];
   int threadParam = 0;  // thread function parameter (not used)
   int i;

   // create threads
   printf("Starting %d threads\n", NUMBER_OF_THREADS);
   for (i = 0; i < NUMBER_OF_THREADS; i++)
      pthread_create(&(tid[i]), NULL, thread, &threadParam);

   // wait for the threads to finish
   for (i = 0; i < NUMBER_OF_THREADS; i++)
      pthread_join(tid[i], NULL);

   printf("\nThreads stopped.\n");

   if (count != (unsigned)ITERATIONS*NUMBER_OF_THREADS)
      printf("A RACE! count=%d (not %d)\n", count, (unsigned)ITERATIONS*NUMBER_OF_THREADS);
   else
      printf("No race: count=%d\n", count);

   return 0;
}//main()



void *thread(void *vargp)
{
   int i;
   for (i = 0; i < ITERATIONS; i++)
   {
    //count++;              // race condition
      asm("incl  _count");  // no race condition (when run on a single core)
   }

   return NULL;
}//thread()